// exception.cc 
//	Entry point into the Nachos kernel from user programs.
//	There are two kinds of things that can cause control to
//	transfer back to here from user code:
//
//	syscall -- The user code explicitly requests to call a procedure
//	in the Nachos kernel.
//
//	exceptions -- The user code does something that the CPU can't handle.
//	For instance, accessing memory that doesn't exist, arithmetic errors,
//	etc.  
//
//	Interrupts (which can also cause control to transfer from user
//	code into the Nachos kernel) are handled elsewhere.
//
// Copyright (c) 1992-1996 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#include "copyright.h"
#
#include "main.h"
#include "syscall.h"
#include "proctable.h"
#include "synchconsole.h"
#include "thread.h"
#include "machine.h"
#include "filesys.h"
#include "addrspace.h"
#include "interrupt.h"
#include <stdio.h>

//--------------------------------------------------------------------
// ProcessStartup
//
//   This is the function run by each new thread when it is Forked.
//   By the time this function is invoked, the new thread should
//   already have an address space and a process identifier.
//   All we need to do is initialize the machine and begin
//   simulation of the user program that is already loaded
//   into the address space
//--------------------------------------------------------------------

static void
ProcessStartup(int dummy)
{
  kernel->currentThread->space->InitRegisters();
  kernel->currentThread->space->RestoreState();
  kernel->machine->Run();
  ASSERT(FALSE);
}

//----------------------------------------------------------------------
// Terminator
//    A helper function to terminate execution of the process that
//    calls it.
//
//    "status" is the desired exit status for the terminating process
//
//    Control never returns from the function
//----------------------------------------------------------------------
void
Terminator(int status)
{
  int id;
  AddrSpace *space;

  id = kernel->currentThread->space->id;

  // tell the process table that this process is done,
  // and record its status value
  kernel->procTable->ProcIsDone(id,status);

  // first zero out the thread's address space pointer
  // from this point on, the thread no longer has an associated
  // address space.  We do this because the actual deletion of
  // the address space may take some time, and we don't want the
  // rest of the kernel to invoke address space methods (via
  // the pointer from this thread) while the address space is
  // being deleted
  space = kernel->currentThread->space;
  kernel->currentThread->space = (AddrSpace *)0;
  // now, delete the address space - this closes any files opened
  // by this process and releases the physical memory it used
  delete space;
  // this will terminate the thread, and use the scheduler to 
  // choose another thread to run
  kernel->currentThread->Finish();
  ASSERT(FALSE);
}

//----------------------------------------------------------------------
// ExecHandler
//    A helper function to implement the Exec system call
//
//    "vaddr" is the address of the filename argument
//
//    returns system call exit status
//----------------------------------------------------------------------
int
ExecHandler(int vaddr)
{
  int rval;
  int childId,myId;
  OpenFile *file;
  char *buf;
  AddrSpace* space;
  Thread* newthread;
  char threadname[MAX_THREAD_NAME_LEN+1];

  myId = kernel->currentThread->space->id;
  buf = new char[MaxStringArgLength];
  if (kernel->currentThread->space->UserStringToKernel(vaddr,buf) < 0) {
    // problem getting the file name from user space
    DEBUG(dbgSysCall,"Error reading file name from user address space");
    rval = -1;
  } else {
    DEBUG(dbgSysCall,"Exec file name is: " << buf );
    // open the file specified by the user
    file = kernel->fileSystem->Open(buf);
    if (file != (OpenFile*)0) {
      // assign an identifier to the new process
      // and get a space for it in the process table
      childId = kernel->procTable->GetNewProcId(myId);
      if (childId >= 0) {
	// create a new address space for the new process
	DEBUG(dbgSysCall,"Exec new process ID is: " << childId);
	space = new AddrSpace(file,kernel->frameMap,childId);
	ASSERT(space != (AddrSpace*)0);
	// check that the address space is OK
	if (space->id == childId) {
	  // build a name for the new thread - for debugging purposes
	  // for now, combine name of executable file and process Id
	  // to make a thread name (which may not be unique)
	  // feel free to change this
	  snprintf(threadname,MAX_THREAD_NAME_LEN+1,"%d(%s)",childId,buf);
	  // create a new thread and assign the address space to it
	  newthread = new Thread(threadname);
	  ASSERT(newthread != (Thread*)0);
	  newthread->space = space;
	  // put the new thread on the ready queue
	  newthread->Fork((VoidFunctionPtr) ProcessStartup, (void *) 0);
	  // we have been successful 
	  // the return value should be the Space ID
	  rval = childId;
	} else {
	  // address space not initialized properly - probably
	  // because of lack of physical memory.  
	  delete space;
	  DEBUG(dbgSysCall,"Error initializing address space - memory full?");
	  rval = -1;
	}
      } else {
	// could not get new process id
	DEBUG(dbgSysCall,"Error obtaining new process ID");
	rval = -1;
      }
      delete file;  // close the executable file
    } else {
      // could not open program file
      DEBUG(dbgSysCall,"Error opening program file");
      rval = -1;
    }
  }
  delete [] buf;
  return(rval);
}

//----------------------------------------------------------------------
// JoinHandler
//    A helper function to implement the Join system call
//
//    "childId" is the id of the process to join
//
//    returns system call exit status
//----------------------------------------------------------------------
int
JoinHandler(int childId)
{
  int rval;
  int myId,status;

  myId = kernel->currentThread->space->id;
  // Join will return the exit status of the child process. If
  // If the child is not finished, the caller will wait until it is.
  // If GetProcStatus returns FALSE, the specified process is
  //  not actually a child of the calling process.
  if (kernel->procTable->GetProcStatus(childId,myId,&status) == TRUE) {
    ASSERT(status >= 0);
    rval = status;
  } else {
    // invalid process id
    DEBUG(dbgSysCall,"Invalid process ID");
    rval = -1;
  }
  return(rval);
}

//----------------------------------------------------------------------
// CreateHandler
//    A helper function to implement the Create system call
//
//    "vaddr" is the address of the filename argument
//
//    returns system call exit status
//----------------------------------------------------------------------
int
CreateHandler(int vaddr)
{
  char *buf;
  int rval;

  buf = new char[MaxStringArgLength];
  // read user-provided file name into kernel buffer
  if (kernel->currentThread->space->UserStringToKernel(vaddr,buf) < 0) {
    // problem gettting the filename from user space
    DEBUG(dbgSysCall,"Error reading file name from user address space");
    rval = -1;
  } else {
    // create the file
#ifdef FILESYS_STUB
    if ( kernel->fileSystem->Create(buf) ) { 
      rval = 0;
    } else {
      DEBUG(dbgSysCall,"Error creating the file");
      rval = -1;
    }
#else
    if ( kernel->fileSystem->Create(buf,0) ) { 
      rval = 0;
    } else {
      DEBUG(dbgSysCall,"Error creating the file");
      rval = -1;
    }
#endif
  }
  delete [] buf;
  return(rval);
}
  
//----------------------------------------------------------------------
// OpenHandler
//    A helper function to implement the Open system call
//
//    "vaddr" is the address of the filename argument
//    "modeaddr" is the address of the string indicating the
//        mode that the file should be opened with
//        e.g., read, read/write, write, append
//        NOTE: this is currently ignored.
//
//    returns system call exit status
//----------------------------------------------------------------------
int
OpenHandler(int vaddr, int modeaddr)
{
  OpenFile* file;
  int fileID;
  char* buf;
       
  buf = new char[MaxStringArgLength];
  // read user-provided file name into kernel buffer
  if (kernel->currentThread->space->UserStringToKernel(vaddr,buf) < 0) {
    // problem reading file name from user space
    DEBUG(dbgSysCall,"Error reading file name from user address space");
    fileID = -1;
  } else {
    // open the file
    file = kernel->fileSystem->Open(buf);
    if (file != (OpenFile*)0) {
      // put the open file into the open file
      // table for this address space
      fileID = kernel->currentThread->space->AddOpenFile(file);
      if (fileID < 0) {
	// there was no room in the open file table
	delete file;
	DEBUG(dbgSysCall,"Error obtaining file ID - process file table full?");
	fileID = -1;
      }
    } else {
      // problem opening the specified file
      DEBUG(dbgSysCall,"Error opening the file");
      fileID = -1;
    }
  }
  delete [] buf;
  return(fileID);
}

//----------------------------------------------------------------------
// ReadHandler
//    A helper function to implement the Read system call
//
//    "fileID" identifies the file to read from
//    "vaddr" is the address of the caller's buffer
//    "length" is the length of the caller's buffer
//
//    returns system call exit status
//----------------------------------------------------------------------
int
ReadHandler(int fileID, int vaddr, int length)
{
    int rval;
    OpenFile* file;
    char* buf;

    if( fileID == ConsoleOutId) {
      // attempt to read from the output console
      // This is an error! 
      DEBUG(dbgSysCall,"Attempt to read from output console.");
      rval = -1;
    } else if (length < 0) {
      DEBUG(dbgSysCall,"Negative read length.");
      rval = -1;
    } else if (length == 0 ) {
      // Nothing to do, but not a device or OS error
      DEBUG(dbgSysCall,"Zero read length.");
      rval = 0;
    } else if (fileID == ConsoleInId) {
      // Read from the input console.
      // Would be better (and more realistic) to read
      // the data in chunks if length is large.
      // However, we don't do that here...
      buf = new char[length];
      // read data from console into kernel buffer
      DEBUG(dbgSysCall,"Input console read.");
      kernel->synchConsoleIn->ConsoleToKernel(length,buf);
      // move data from kernel buffer to user virtual address space
      if (kernel->currentThread->space->KernelToUserBuf(vaddr,length,buf) 
	  != length) {
	DEBUG(dbgSysCall,"Error copying data to user address space");
	rval = -1;
      } else {
	rval = length;
      }
      delete [] buf;
    } else {
      // Read from a file...
      // find the open file that corresponds to the file ID
      // passed to the kernel by the user process
      DEBUG(dbgSysCall,"File read.");
      file = kernel->currentThread->space->GetOpenFile(fileID);
      if (file != (OpenFile*)0) {
	// if we get here, we were passed a legit file ID
	// everything looks good - time to read the data
	buf = new char[length];
	// read data from the file into the user buffer
	// rval is the number of bytes actually read
	// Would be better (and more realistic) to read
	// the data in chunks if length is large.
	// However, we don't do that here...
	rval = file->Read(buf,length);
	ASSERT(rval >= 0 && rval <= length);
	// copy data from kernel buffer to user buffer
	if (kernel->currentThread->space->KernelToUserBuf(vaddr,rval,buf)
	    != rval) {
	  DEBUG(dbgSysCall,"Error copying data to user address space");
	  rval = -1;
	}
	delete [] buf;
      } else {
	// problem with fileID supplied by caller
	DEBUG(dbgSysCall,"Invalid file ID.");
	rval = -1;
      }
    }
    return(rval);
}

//----------------------------------------------------------------------
// WriteHandler
//    A helper function to implement the Write system call
//
//    "fileID" identifies the file to write from
//    "vaddr" is the address of the caller's buffer
//    "length" is the length of the caller's buffer
//
//    returns system call exit status
//----------------------------------------------------------------------
int
WriteHandler(int fileID, int vaddr, int length)
{
    int rval;
    OpenFile* file;
    char* buf;

    if (fileID == ConsoleInId) {
      // Write to the console input
      // this is an error
      DEBUG(dbgSysCall,"Attempt to write to input console.");
      rval = -1;
    } else if (length < 0) {
      DEBUG(dbgSysCall,"Negative write length.");
      rval = -1;
    } else if (length == 0) {
      // Nothing to do, but not a system or OS error
      DEBUG(dbgSysCall,"Zero write length.");
      rval = 0;
    } else if (fileID == ConsoleOutId) {
      // Write to the console...
      buf = new char[length];
      // get the user string into a buffer
      // Would be better (and more realistic) to copy
      // the data in chunks if length is large.
      // However, we don't do that here...
      DEBUG(dbgSysCall,"Output console write.");
      if (kernel->currentThread->space->UserBufToKernel(vaddr,length,buf)
	  != length) {
	// problem reading in user's buffer
	DEBUG(dbgSysCall,"Error copying data from user address space");
	rval = -1;
      } else {
	// write the buffer contents to the console
	kernel->synchConsoleOut->KernelToConsole(length,buf);
	rval = length;
      }
      delete [] buf;
    } else {
      // Write to a file...
      // find the open file that corresponds to the file ID
      // passed to the kernel by the user process
      DEBUG(dbgSysCall,"File write.");
      file = kernel->currentThread->space->GetOpenFile(fileID);
      if (file != (OpenFile*)0) {
	// if we get here, we were passed a legit file ID
	// everything looks good - time to write the data
	buf = new char[length];
	// get the user data into a buffer
	// Would be better (and more realistic) to copy
	// the data in chunks if length is large.
	// However, we don't do that here...
	if (kernel->currentThread->space->UserBufToKernel(vaddr,length,buf)
	    != length) {
	  // problem reading in the user's buffer
	  DEBUG(dbgSysCall,"Error copying data from user address space");
	  rval = -1;
	} else {
	  // write the data to the file
	  rval = file->Write(buf,length);
	  ASSERT(rval >= 0 && rval <= length);
	}
	delete [] buf;
      } else {
	// problem with fileID passed by caller
	DEBUG(dbgSysCall,"Invalid file ID.");
	rval = -1;
      }
    }
    return(rval);
}
      
//----------------------------------------------------------------------
// CloseHandler
//    A helper function to implement the Close system call
//
//    "fileID" identifies the file to close
//
//    returns system call exit status
//----------------------------------------------------------------------
int
CloseHandler(int fileID)
{
  int rval;
  OpenFile* file;

  // first, find the open file on the open file list
  // and remove it from the list
  file = kernel->currentThread->space->DeleteOpenFile(fileID);
  // if we found the open file, close it
  if (file != (OpenFile*)0) {
    delete file;
    rval = 0;
  } else {
    DEBUG(dbgSysCall,"Invalid file ID.");
    rval = -1;
  }
  return(rval);
}

//----------------------------------------------------------------------
// ExceptionHandler
// 	Entry point into the Nachos kernel.  Called when a user program
//	is executing, and either does a syscall, or generates an addressing
//	or arithmetic exception.
//
// 	For system calls, the following is the calling convention:
//
// 	system call code -- r2
//		arg1 -- r4
//		arg2 -- r5
//		arg3 -- r6
//		arg4 -- r7
//
//	The result of the system call, if any, must be put back into r2. 
//
//      If you are handling a system call, don't forget to increment the pc
//      before returning. (Or else you'll loop making the same system call
//      forever!)
//
//	"which" is the kind of exception.  The list of possible exceptions 
//	is in machine.h.
//----------------------------------------------------------------------

void
ExceptionHandler(ExceptionType which)
{
    int type = kernel->machine->ReadRegister(2);
    int fileID;  
    int vaddr;
    int modeaddr;
    int length;
    int returnval;
    int childId;
    int status;

    ASSERT(kernel->currentThread != (Thread*)0);
    ASSERT(kernel->currentThread->space != (AddrSpace*)0);
    // default system call return value
    returnval = 0;

    switch(which) {
    case SyscallException:
      switch(type) {

      case SC_Halt:
	DEBUG(dbgSysCall,"System Call: Halt");
        kernel->interrupt->Halt();
        ASSERT(FALSE);

      case SC_Exit:
        status = kernel->machine->ReadRegister(4);
        // we require the status value to be non-negative, since
        // negative values are used to indicate errors during Join
        // If the status is negative, make it positive.
	DEBUG(dbgSysCall,"System Call: Exit status=" << status);
        if (status < 0) status *= (-1);
        ASSERT(status >= 0);
        Terminator(status);
        break;

      case SC_Exec:
        vaddr = kernel->machine->ReadRegister(4);
	DEBUG(dbgSysCall,"System Call: Exec vaddr=" << vaddr);
        returnval = ExecHandler(vaddr);
        break;

      case SC_Join:
        childId = kernel->machine->ReadRegister(4);
	DEBUG(dbgSysCall,"System Call: Join pid=" << childId);
	returnval = JoinHandler(childId);
        break;

      case SC_Create:
        vaddr = kernel->machine->ReadRegister(4);
	DEBUG(dbgSysCall,"System Call: Create vaddr=" << vaddr);
	returnval = CreateHandler(vaddr);
        break;

      case SC_Open:
        vaddr = kernel->machine->ReadRegister(4);
        modeaddr = kernel->machine->ReadRegister(5);
	DEBUG(dbgSysCall,"System Call: Open vaddr=" << vaddr);
	DEBUG(dbgSysCall,"System Call: Open modeaddr=" << modeaddr);
	returnval = OpenHandler(vaddr, modeaddr);
        break;

      case SC_Read:
        fileID = kernel->machine->ReadRegister(6);
        vaddr = kernel->machine->ReadRegister(4);
        length = kernel->machine->ReadRegister(5);
	DEBUG(dbgSysCall,"System Call: Read vaddr=" << vaddr << " length=" << length << " fileID=" << fileID);
	returnval = ReadHandler(fileID,vaddr,length);
        break;

      case SC_Write:
        fileID = kernel->machine->ReadRegister(6);
        vaddr = kernel->machine->ReadRegister(4);
        length = kernel->machine->ReadRegister(5);
	DEBUG(dbgSysCall,"System Call: Write vaddr=" << vaddr << " length=" << length << " fileID=" << fileID);
	returnval = WriteHandler(fileID,vaddr,length);
        break;

      case SC_Close:
        fileID = kernel->machine->ReadRegister(4);
	DEBUG(dbgSysCall,"System Call: Close fileID=" << fileID);
	returnval = CloseHandler(fileID);
        break;

      default:
	// remaining system calls are not implemented
	DEBUG(dbgSysCall,"Unknown or Unimplemented System Call!");
	// we simply do nothing
	break;
      }
      break;

    case BusErrorException:
      DEBUG(dbgSysCall,"bus error exception, invalid address.");
      Terminator(1);
      break;

    case OverflowException:
      DEBUG(dbgSysCall,"overflow exception.");
      Terminator(1);
      break;

    case IllegalInstrException:
      DEBUG(dbgSysCall,"illegal instruction exception.");
      Terminator(1);
      break;

    default:
      DEBUG(dbgSysCall,"Unknown or unimplemented exception!\n");
      Terminator(1);
      break;
    }

    // if this was a system call, 
    //  place the return value in r2, and
    // increment the PC before returning
    if (which == SyscallException) {
      kernel->machine->WriteRegister(PCReg,kernel->machine->ReadRegister(PCReg)+4);
      kernel->machine->WriteRegister(NextPCReg,kernel->machine->ReadRegister(NextPCReg)+4);
      // we always place a return value into the register,
      // even if the system call does not require one.
      // this just simplifies the code....
      kernel->machine->WriteRegister(2,returnval);
    }
    return;
}



